VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
END
Attribute VB_Name = "Telnet"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False

'basiert auf Public-Domain-Code von Ian Storrs
'http://storrs.home.ml.org/video/telnet.zip
'storrs@bellsouth.net

Option Explicit

Public Terminal As String
Public Echo As Boolean
Public Trace As Boolean

Private Const LinesPerPage = 25
Private Const CharsPerLine = 80

Private WithEvents Socket As CSocket
Attribute Socket.VB_VarHelpID = -1
Event Connected(Host$)
Event Closed()
Event Cursor(visible As Boolean)
Event CharArrival(ch As Byte)
Private Receiving As Boolean

Private Enum States
   FSM_iac_init = 1
   FSM_Telnet_iac_first = 2
   FSM_Telnet_iac_do = 3
   FSM_Telnet_iac_dont = 4
   FSM_Telnet_iac_will = 5
   FSM_Telnet_iac_wont = 6
   FSM_Telnet_iac_sb = 7
   FSM_terminal = 8
End Enum

Private Const SUSP = 237
Private Const ABORT = 238      'Abort
Private Const SE = 240         'End of Subnegotiation
Private Const NOP = 241
Private Const DM = 242         'Data Mark
Private Const BREAK = 243      'BREAK
Private Const IP = 244         'Interrupt Process
Private Const AO = 245         'Abort Output
Private Const AYT = 246        'Are you there
Private Const EC = 247         'Erase character
Private Const EL = 248         'Erase Line
Private Const GOAHEAD = 249    'Go Ahead
Private Const SB = 250         'What follows is subnegotiation
Private Const WILLTEL = 251
Private Const WONTTEL = 252
Private Const DOTEL = 253
Private Const DONTTEL = 254
Private Const IAC = 255

Private Const Telnet_IS = 0
Private Const Telnet_SEND = 1

Private Const BINARY = 0
Private Const SENDECHO = 1
Private Const RECONNECT = 2
Private Const SGA = 3
Private Const AMSN = 4
Private Const Status = 5
Private Const TIMING = 6
Private Const RCTAN = 7
Private Const OLW = 8
Private Const OPS = 9
Private Const OCRD = 10
Private Const OHTS = 11
Private Const OHTD = 12
Private Const OFFD = 13
Private Const OVTS = 14
Private Const OVTD = 15
Private Const OLFD = 16
Private Const XASCII = 17
Private Const LOGOUT = 18
Private Const BYTEM = 19
Private Const DET = 20
Private Const SUPDUP = 21
Private Const SUPDUPOUT = 22
Private Const SENDLOC = 23
Private Const TERMTYPE = 24
Private Const EOR = 25
Private Const TACACSUID = 26
Private Const OUTPUTMARK = 27
Private Const TERMLOCNUM = 28
Private Const REGIME3270 = 29
Private Const X3PAD = 30
Private Const NAWS = 31
Private Const TERMSPEED = 32
Private Const TFLOWCNTRL = 33
Private Const LINEMODE = 34
Private Const DISPLOC = 35
Private Const ENVIRON = 36
Private Const AUTHENTICATION = 37
Private Const ENVIRONMENT = 39
Private Const CHARSET = 42
Private Const EXTENDED_OPTIONS_LIST = 255
Private Const RANDOM_LOSE = 256

Private Telnet_State         As States
Private Telnet_SubState      As Boolean
Private Telnet_SB_Data(10)   As Integer
Private Telnet_SB_Index      As Integer
Private Telnet_sw_ugoahead   As Boolean
Private Telnet_sw_igoahead   As Boolean

Private Sub Class_Initialize()
Set Socket = New CSocket
Terminal = "VT100"
Echo = True
Trace = False
End Sub

Private Sub Class_Terminate()
Socket.Disconnect
Set Socket = Nothing
End Sub

Public Function Connect(Host$) As Boolean

If Socket.State <> sockClosed Then
   Socket.Disconnect
   Do: DoEvents: Loop Until Socket.State = sockClosed
End If

Socket.Connect Host$, 23
Do: DoEvents: Loop While Socket.State = sockConnecting

Connect = Socket.State = sockConnected

End Function

Public Sub Disconnect()

If Socket.State = sockConnected Then
   RaiseEvent Closed
   Socket.Disconnect
End If

End Sub

Public Sub SendData(data$)

If Socket.State = sockConnected Then
   Socket.SendData data$
End If

End Sub

Private Sub Socket_Closed()

RaiseEvent Closed
Socket.Disconnect

End Sub

Private Sub Socket_Connect()

Telnet_sw_ugoahead = True
Telnet_sw_igoahead = False

Telnet_SubState = False
Telnet_State = FSM_iac_init

If Echo Then Telnet_SendDo SENDECHO Else Telnet_SendWont SENDECHO
Telnet_SendDo SGA
Telnet_SendWill NAWS
If Len(Terminal) Then Telnet_SendWill TERMTYPE
Telnet_SendWill TERMSPEED

RaiseEvent Connected(Socket.RemoteHost)

End Sub

Private Sub Socket_DataArrival(BinData() As Byte, RemoteHostIP As String)

Dim ch()     As Byte
Dim I        As Integer

If Receiving Then Exit Sub

Receiving = True
RaiseEvent Cursor(False)

For I = LBound(BinData) To UBound(BinData)
   Telnet_ProcessChar BinData(I)
Next I

RaiseEvent Cursor(True)
Receiving = False

End Sub

Private Sub SOCKET_ERROR(Error As Long, Description As String)

Socket_Closed

End Sub

Private Function Telnet_iac_first(ch As Byte) As Integer

Select Case ch
Case DOTEL
   Telnet_iac_first = FSM_Telnet_iac_do
Case DONTTEL
   Telnet_iac_first = FSM_Telnet_iac_dont
Case WILLTEL
   Telnet_iac_first = FSM_Telnet_iac_will
Case WONTTEL
   Telnet_iac_first = FSM_Telnet_iac_wont
Case SB
   Telnet_iac_first = FSM_Telnet_iac_sb
   Telnet_SB_Index = 0
Case SE
   If Trace Then Debug.Print "NEGOTIATION ENDED "
   Telnet_iac_first = FSM_terminal
Case EOR
   If Trace Then Debug.Print "EOR"
   Telnet_iac_first = FSM_terminal
Case IAC
   RaiseEvent CharArrival(ch)
   Telnet_iac_first = FSM_terminal
Case Else
   If Trace Then Debug.Print "ERROR in state IAC " & ch
   Telnet_iac_first = FSM_terminal
End Select

End Function

Private Function Telnet_iac_do(ch As Byte) As Integer

'DO Processing Respond with WILL or WONT

Telnet_iac_do = FSM_terminal
If ch = SE Then Exit Function 'NEGOTIATION ENDED

If Trace Then Debug.Print "RECEIVED: DO "; Telnet_FlagName$(ch)

Select Case ch
Case NAWS
   Telnet_SendNegotiation ch, Chr$(CharsPerLine) & Chr$(0) & Chr$(LinesPerPage), "80x24"
Case SGA
   If Not Telnet_sw_igoahead Then
      Telnet_SendWill ch
      Telnet_sw_igoahead = True
   End If
Case TERMTYPE
   If Len(Terminal) > 0 Then
      Telnet_SendWill ch
      Telnet_SendNegotiation ch, Terminal, Terminal
   Else
      Telnet_SendWont ch
   End If
Case TERMSPEED
   Telnet_SendWill ch
   Telnet_SendNegotiation ch, "57600,57600", "57600"
Case AUTHENTICATION
   Telnet_SendWill ch
   Telnet_SendNegotiation ch, Chr$(0) & Chr$(0) & Chr$(0), ""
Case Else
   Telnet_SendWont ch
End Select

End Function

Private Function Telnet_iac_dont(ch As Byte) As Integer

'DONT Processing

Telnet_iac_dont = FSM_terminal
If ch = SE Then Exit Function 'NEGOTIATION ENDED

If Trace Then Debug.Print "RECEIVED: DONT "; Telnet_FlagName$(ch)

Select Case ch
Case SGA
   If Not Telnet_sw_ugoahead Then
      Telnet_SendWont ch
      Telnet_sw_ugoahead = True
   End If
Case Else
   Telnet_SendWont ch
End Select

End Function

Private Function Telnet_iac_will(ch As Byte) As Integer

' WILL Processing - Respond with DO or DONT

Telnet_iac_will = FSM_terminal
If ch = SE Then Exit Function 'NEGOTIATION ENDED

If Trace Then Debug.Print "RECEIVED: WILL "; Telnet_FlagName$(ch)

Select Case ch
Case SENDECHO
   If Echo Then Telnet_SendDo ch Else Telnet_SendDont ch
Case SGA
   If Not Telnet_sw_ugoahead Then
      Telnet_SendDo ch
      Telnet_sw_ugoahead = True
   End If
Case Else
   Telnet_SendDont ch
End Select

End Function

Private Function Telnet_iac_wont(ch As Byte) As Integer

' WONT Processing

Telnet_iac_wont = FSM_terminal
If ch = SE Then Exit Function 'NEGOTIATION ENDED

If Trace Then Debug.Print "RECEIVED: WONT "; Telnet_FlagName$(ch)

Select Case ch
Case SENDECHO
   If Echo Then Telnet_SendDo ch Else Telnet_SendDont ch
Case SGA
   'Telnet_SendDont ch
   Telnet_sw_igoahead = False
Case Else
   'Telnet_SendDont ch
End Select

End Function

Private Function Telnet_iac_sb(ch As Byte) As Integer

' Collect parms after SB and until another IAC

Select Case ch
Case IAC
   Telnet_iac_sb = FSM_Telnet_iac_sb

Case SE
   ' End of negotiation string, string is in Telnet_SB_Data()
   Telnet_iac_sb = FSM_terminal
   If Trace Then Debug.Print "RECEIVED: SE "
   Select Case Telnet_SB_Data(0)
   Case TERMTYPE
      If Telnet_SB_Data(1) = Telnet_SEND Then
         Telnet_SendWill TERMTYPE
         Telnet_SendNegotiation TERMTYPE, Terminal, Terminal
         'If Trace Then Debug.Print "SENT: SB TERMTYPE VT100"
         'Socket.SendData Chr$(IAC) & Chr$(SB) & Chr$(TERMTYPE) & "VT100" & Chr$(0) & Chr$(IAC) & Chr$(SE)
      End If
   Case TERMSPEED
      If Telnet_SB_Data(1) = Telnet_SEND Then
         Telnet_SendWill TERMSPEED
         Telnet_SendNegotiation TERMSPEED, "57600,57600", "57600"
      End If
   End Select

Case Else
   If Trace Then Debug.Print "RECEIVED: SB("; Telnet_SB_Index; ") = " & ch

   Telnet_SB_Data(Telnet_SB_Index) = ch
   Telnet_SB_Index = Telnet_SB_Index + 1

   Telnet_iac_sb = FSM_Telnet_iac_sb
End Select

End Function

Private Sub Telnet_ProcessChar(ch As Byte)

Select Case Telnet_State
Case FSM_terminal
   If ch = IAC Then
      Telnet_State = FSM_Telnet_iac_first
   Else
      RaiseEvent CharArrival(ch)
   End If
Case FSM_Telnet_iac_first
   Telnet_State = Telnet_iac_first(ch)
Case FSM_Telnet_iac_do
   Telnet_State = Telnet_iac_do(ch)
Case FSM_Telnet_iac_dont
   Telnet_State = Telnet_iac_dont(ch)
Case FSM_Telnet_iac_will
   Telnet_State = Telnet_iac_will(ch)
Case FSM_Telnet_iac_wont
   Telnet_State = Telnet_iac_wont(ch)
Case FSM_Telnet_iac_sb
   Telnet_State = Telnet_iac_sb(ch)
Case FSM_iac_init
   If ch = IAC Then
      Telnet_State = FSM_Telnet_iac_first
   Else
      RaiseEvent CharArrival(ch)
   End If
Case Else
   If Trace Then Debug.Print "Invalid 'next (" + Str$(Telnet_State) + ")' processing routine in Telnet_State loop"
End Select

End Sub

Private Function Telnet_FlagName$(ch As Byte)

Select Case ch
Case BINARY:         Telnet_FlagName$ = "BINARY"
Case SENDECHO:       Telnet_FlagName$ = "ECHO"
Case NAWS:           Telnet_FlagName$ = "WINDOW SIZE"
Case SGA:            Telnet_FlagName$ = "SGA"
Case TERMTYPE:       Telnet_FlagName$ = "TERMTYPE"
Case TERMSPEED:      Telnet_FlagName$ = "TERMSPEED"
Case TFLOWCNTRL:     Telnet_FlagName$ = "FLOWCONTROL"
Case LINEMODE:       Telnet_FlagName$ = "LINEMODE"
Case Status:         Telnet_FlagName$ = "STATUS"
Case TIMING:         Telnet_FlagName$ = "TIMING"
Case DISPLOC:        Telnet_FlagName$ = "DISPLOC"
Case ENVIRON:        Telnet_FlagName$ = "ENVIRON"
Case ENVIRONMENT:    Telnet_FlagName$ = "ENVIRONMENT"
Case AUTHENTICATION: Telnet_FlagName$ = "AUTHENTICATION"
Case Else:           Telnet_FlagName$ = "UNKNOWN CMD " & Asc(ch)
End Select

End Function

Private Sub Telnet_SendWont(ch As Byte)
If Trace Then Debug.Print "SENT: WONT "; Telnet_FlagName$(ch)
Socket.SendData Chr$(IAC) & Chr$(WONTTEL) & Chr$(ch)
End Sub

Private Sub Telnet_SendWill(ch As Byte)
If Trace Then Debug.Print "SENT: WILL "; Telnet_FlagName$(ch)
Socket.SendData Chr$(IAC) & Chr$(WILLTEL) & Chr$(ch)
End Sub

Private Sub Telnet_SendDo(ch As Byte)
If Trace Then Debug.Print "SENT: DO "; Telnet_FlagName$(ch)
Socket.SendData Chr$(IAC) & Chr$(DOTEL) & Chr$(ch)
End Sub

Private Sub Telnet_SendDont(ch As Byte)
If Trace Then Debug.Print "SENT: DONT "; Telnet_FlagName$(ch)
Socket.SendData Chr$(IAC) & Chr$(DONTTEL) & Chr$(ch)
End Sub

Private Sub Telnet_SendNegotiation(ch As Byte, par$, debugcaption$)
If Trace Then Debug.Print "SENT: SB "; Telnet_FlagName$(ch); " "; debugcaption$; " SE"
Socket.SendData Chr$(IAC) & Chr$(SB) & Chr$(ch) & Chr$(0) & par$ & Chr$(IAC) & Chr$(SE)
End Sub

